{# Project list header #}
{# Expects to have custom_cols set #}
{% load projectroles_common_tags %}
{% get_django_setting 'PROJECTROLES_DISABLE_CATEGORIES' as disable_categories %}

<tr class="sodar-pr-project-list-header">
  <th id="sodar-pr-project-list-header-project">
    {% get_display_name 'PROJECT' title=True %}
    {% if not disable_categories %}
      / {% get_display_name 'CATEGORY' title=True %}
    {% endif %}
  </th>
  {% for col in project_custom_cols %}
    {% if col.active %}
      <th id="sodar-pr-project-list-header-{{ col.app_plugin.name }}-{{ col.key }}"
          class="text-nowrap {% if col.align %}text-{{ col.align }}{% endif %}"
        {% if col.width %} style="width: {{ col.width }}px !important"{% endif %}>
        {{ col.title }}
        {% if col.description %}
          {% get_info_link col.description as info_link %}
          {{ info_link | safe }}
        {% endif %}
      </th>
   {% endif %}
 {% endfor %}
  <th id="sodar-pr-project-list-header-role">Your&nbsp;Role</th>
</tr>
