{% extends 'projectroles/project_base.html' %}

{% load rules %}
{% load projectroles_tags %}

{% block title %}
  Confirm Removal of Member {{ object.user.username }} from {{ object.project.title }}
{% endblock title %}

{% block projectroles_extend %}

<div class="container-fluid sodar-page-container">
  <h3>Confirm Removal of Member {{ object.user.username }}</h3>

  <div class="alert alert-warning" role="alert">
    Do you really want to remove the role of "{{ object.role.name }}"
    from user {{ object.user.username }} in {{ object.project.title }}?
  </div>

  <form method="post">
    {% csrf_token %}
    <div class="btn-group pull-right">
      <a role="button" class="btn btn-secondary"
          href="{{ request.session.real_referer }}">
        <i class="iconify" data-icon="mdi:arrow-left-circle"></i> Cancel
      </a>
      <button type="submit" class="btn btn-danger">
        <i class="iconify" data-icon="mdi:account-remove"></i> Remove
      </button>
    </div>
  </form>
</div>

{% endblock projectroles_extend %}
